#include <unistd.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "accounts.h"
#include "configfile.h"

#define CONFIG_FILE "/.network.conf"

#if 1
#define PEER_FILE  "/etc/ppp/peers/network"
#define CHATSCRIPT_FILE  "/etc/ppp/chatscripts/network"
#define RESOLV_FILE    "/etc/ppp/resolv/network"
#define PAP_SECRETS_FILE  "/etc/ppp/pap-secrets"
#define CHAP_SECRETS_FILE "/etc/ppp/chap-secrets"
#define config_dir "/home/default/.network/"
#else
#define CHATSCRIPT_FILE "/tmp/chatscript"
#define PEER_FILE "/tmp/network"
#define PAP_SECRETS_FILE "/tmp/pap-secrets"
#define CHAP_SECRETS_FILE "/tmp/chap-secrets"
#define config_dir "/home/james/.network/"
#endif

#ifndef OPEN_MAX
#define OPEN_MAX 12000
#endif

char *optcmp(const char *opt, char *buf);
int getbool(const char *s);
int device(const char *s);
static void write_chatscript(Account *acct);
static void write_pap_secrets(Account *acct);
static void write_chap_secrets(Account *acct);
static void write_peer(Account *acct);
void sigchld_handler(int);

static char *conffile;

Account_list *al_load(void)
{
	Account_list *al;
	char *home;
	char *names, *p;
	int i;

	home = getenv ("HOME");
	conffile = malloc(strlen (home) + strlen (CONFIG_FILE) + 1);
	if (home && conffile)
	{
		strcpy (conffile, home);
		strcat (conffile, CONFIG_FILE);
	}
	printf("%s\n", conffile);
	configfile_parse(conffile);

	al = calloc(1, sizeof(Account_list));
	names = strdup(get_param_str("general", "accounts", NULL));
	for(p=names; *p; p++) 
		if(*p==' ') al->n_accts++;
	if(*names) al->n_accts++;

	al->accts = calloc(al->n_accts, sizeof(Account));
	for(i=0;i<al->n_accts;i++) {
		char buf[256];
		FILE *fp;

		al->accts[i].section = strdup(strtok(i ? NULL : names, " "));
		strcpy(buf, "/tmp/network_");
		strcat(buf, al->accts[i].section);
		strcat(buf, ".pid");

		if( (fp=fopen(buf, "r")) != NULL) {
			fscanf(fp, "%d", &al->accts[i].pid);
			fclose(fp);
		} else
			al->accts[i].pid=0;
	}
	free(names);
	return al;
}

int al_new_acct(Account_list *al)
{
	char *s, tmp[20];
	const char *p;
	int i, j;

	al->accts = realloc(al->accts, sizeof(Account) * ++al->n_accts);
	memset(al->accts + al->n_accts - 1, 0, sizeof(Account));

	p = get_param_str("general", "accounts", "");
	i = strlen(p);
	s = malloc(i + 20);
	strcpy(s, p);

	for(j=1;j<=al->n_accts;j++) {
		sprintf(tmp, "acct%d", j);
		if(!strstr(p, tmp)) break;
	}
	strcat(s, " ");
	strcat(s, tmp);
	set_param_str("general", "accounts", s);
	al->accts[al->n_accts-1].section = strdup(tmp);
	free(s);

	return al->n_accts - 1;
}

int al_delete_acct(Account_list *al, int i)
{
	char *s;
	int len=0;
	al->n_accts--;

	delete_section(al->accts[i].section);
	free(al->accts[i].section);
	if(i!=al->n_accts) 
		memcpy(al->accts+i, al->accts+al->n_accts, sizeof(Account));

	for(i=0;i<al->n_accts;i++)
		len += strlen(al->accts[i].section);
	s=malloc(len + i + 1);
	*s=0;
	for(i=0;i<al->n_accts;i++) {
		strcat(s, al->accts[i].section);
		if(i!=al->n_accts-1) strcat(s, " ");
	}
	set_param_str("general", "accounts", s);
	free(s);

	acct_save();
	return 0;
}

void acct_save(void)
{
	configfile_write(conffile);
}

const char *acct_name(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "name", "");
}

const char *acct_type(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "type", "DIRECT");
}

const char *acct_fctrl(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "flow_control", "CRTSCTS");
}

const char *acct_dnstype(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "dnstype", "DYNAMIC");
}

const char *acct_device(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "device", "/dev/ttyS0");
}

const char *acct_phone(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "phone", "");
}

const char *acct_user(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "username", "");
}

const char *acct_pass(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "password", "");
}

int acct_defaultgw(Account_list *al, int i)
{
	if(!strcmp(get_param_str(al->accts[i].section,
				 "default_gateway", "YES"), "YES"))
		return 1;
	else return 0;
}

int acct_hcomp(Account_list *al, int i)
{
	if(!strcmp(get_param_str(al->accts[i].section,
				 "header_compression", "YES"), "YES"))
		return 1;
	else return 0;
}

const char *acct_speed(Account_list *al, int i)
{
	return get_param_str(al->accts[i].section, "speed", "115200");
}

void acct_set_name(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "name", s);
}

void acct_set_type(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "type", s);
}

void acct_set_dnstype(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "dnstype", s);
}

void acct_set_fctrl(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "flow_control", s);
}

void acct_set_device(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "device", s);
}

void acct_set_phone(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "phone", s);
}

void acct_set_user(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "username", s);
}

void acct_set_pass(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "password", s);
}

void acct_set_speed(Account_list *al, int i, char *s)
{
	set_param_str(al->accts[i].section, "speed", s);
}

void acct_set_defaultgw(Account_list *al, int i, int val)
{
	set_param_str(al->accts[i].section, "default_gateway", 
		      val ? "YES" : "NO");
}

void acct_set_hcomp(Account_list *al, int i, int val)
{
	set_param_str(al->accts[i].section, "header_compression", 
		      val ? "YES" : "NO");
}

void acct_sethandle(Account_list *al, int i, pghandle h)
{
	al->accts[i].handle = h;
}

pghandle acct_gethandle(Account_list *al, int i)
{
	return al->accts[i].handle;
}

int acct_isconnected(Account_list *al, int i)
{
	if(!al->accts[i].pid) return 0;
	return(!kill(al->accts[i].pid, 0));
}

int acct_start(Account_list *al, int i)
{
	pid_t pid;
	int fd, pfd_[2], pfderr_[2];
	char buf[256], pfdstr[10];
	FILE *fp;

	write_peer(&al->accts[i]);
	write_chatscript(&al->accts[i]);
	write_pap_secrets(&al->accts[i]);
	write_chap_secrets(&al->accts[i]);

	signal(SIGCHLD, SIG_IGN);

	if(pipe(pfd_) == -1) {
		fprintf(stderr, "pipe err\n");
		exit(1);
	}
	if(pipe(pfderr_) == -1) {
		fprintf(stderr, "pipe err\n");
		exit(1);
	}

	pid = fork();
	if(pid == 0) {
		if(setpgrp()==-1) exit(0);
		for (fd = 3; fd < OPEN_MAX; fd++) 
			if((fd != pfd_[1]) && fd > 2 && (fd != pfderr_[1]))
				close(fd);

		close(2);
		if(dup(pfderr_[1]) != 2) exit(1);
		close(pfderr_[1]);

		snprintf(pfdstr, 9, "%d", pfd_[1]);
		execl("/usr/sbin/pppd", "pppd", "logfd", pfdstr,
		      "call", "network", (void* )0);
	}
	al->accts[i].pid=pid;
	close(pfd_[1]);
	close(pfderr_[1]);
	al->accts[i].pfd=pfd_[0];
	al->accts[i].pfderr=pfderr_[0];

	strcpy(buf, "/tmp/network_");
	strcat(buf, al->accts[i].section);
	strcat(buf, ".pid");
	if((fp=fopen(buf, "w"))) {
		fprintf(fp, "%d", pid);
		fclose(fp);
	} else
		perror("fopen");

	return 0;
}

int acct_stop(Account_list *al, int i)
{
	int j;
	char buf[256];
	
	/*Give 10 seconds to die*/
	if(!al->accts[i].pid) return -1;
	for(j=0;j<10 && !kill(al->accts[i].pid, SIGTERM); j++)
		sleep(1);

	strcpy(buf, "/tmp/network_");
	strcat(buf, al->accts[i].section);
	strcat(buf, ".pid");
	unlink(buf);
	return j==10 ? -1 : 0;
}

static void write_chatscript(Account *acct)
{
	FILE *fp;
	const char *s;

	fp=fopen(CHATSCRIPT_FILE, "w");

	s = get_param_str(acct->section, "type", "DIRECT");
	if(!strcmp(s, "MODEM")) {
		fprintf(fp,
			"ABORT BUSY\n"
			"ABORT 'NO CARRIER'\n"
			"ABORT VOICE\n"
			"ABORT 'NO DIALTONE'\n"
			"ABORT 'NO DIAL TONE'\n"
			"ABORT 'NO ANSWER'\n");

		s = get_param_str(acct->section, "init_string", "ATZ");
		fprintf(fp, "'' '%s'\n", s);
		fprintf(fp, "OK-AT-OK 'ATDT%s'\n",
			get_param_str(acct->section, "phone", ""));
		fprintf(fp, "CONNECT \\d\\c\n");
	}
	fclose(fp);
}

static void write_peer(Account *acct)
{
	FILE *fp;
	const char *s;

	fp=fopen(PEER_FILE, "w");

	s = get_param_str(acct->section, "type", "DIRECT");
	if(!strcmp(s, "MODEM"))
		fprintf(fp, "hide-password\n");

	fprintf(fp, "noauth\n");
	fprintf(fp, "debug\n");

	s = get_param_str(acct->section, "device", "/dev/ttyS0");
	fprintf(fp, "%s\n", s);

	if(!strcmp(s, "MODEM")) fprintf(fp, "modem\n");
	else if(!strcmp(s, "DIRECT")) fprintf(fp, "local\n");

	s = get_param_str(acct->section, "username", "");
	if(*s) {
		fprintf(fp, "user %s\n", s);
		fprintf(fp, "remotename network\n");
	}
	fprintf(fp, "ipparam network\n");

	s = get_param_str(acct->section, "dnstype", "NONE");
	if(!strcmp(s, "DYNAMIC"))
		fprintf(fp, "usepeerdns\n");
	else { 
#if 0
		if(iptype == STATIC) {
			if (!assign_remote_ip)
				peer_options << ip << ":\n";
			else {
				char buf[32];
				sprintf(buf, ":%d.%d.%d.%d\n", ip.value0(), ip.value1(), ip.value2(), ip.value3() + 1);
				peer_options << ip << buf;
			}
		}
#endif
	}

	fprintf(fp, "lock\n");
	fprintf(fp, "nodetach\n");

	fprintf(fp, "connect \"/usr/sbin/chat -v -s -S -t60 -f "
		CHATSCRIPT_FILE "\"\n");

	if(!strcmp(get_param_str(acct->section, "header_compression", "NO"),
		   "NO"))
		fprintf(fp, "novj\n");

	if(!strcmp(get_param_str(acct->section, "default_gateway", "NO"),
		   "NO"))
		fprintf(fp, "nodefaultroute\n");
	else
		fprintf(fp, "defaultroute\n");

	fprintf(fp, "%d\n", get_param_int(acct->section, "speed", 115200));

	s = get_param_str(acct->section, "flow_control", "CRTSCTS");
	if(!strcmp(s, "CRTSCTS"))
		fprintf(fp, "crtscts\n");
	else if(!strcmp(s, "CDTRCTS"))
		fprintf(fp, "cdtrcts\n");
	else if(!strcmp(s, "XONXOFF"))
		fprintf(fp, "xonoff\n");
	else if(!strcmp(s, "NONE"))
		fprintf(fp, "nocrtscts\n");

	fclose(fp);
}

static void write_pap_secrets(Account *acct)
{
	FILE *fp;
	const char *username, *password;
  
	username = get_param_str(acct->section, "username", "");
	password = get_param_str(acct->section, "password", "");
	if(*username) {
		fp=fopen(PAP_SECRETS_FILE, "w");

		fprintf(fp, "%s network \"%s\"\n", username, password);
		fprintf(fp, "network %s \"%s\"\n", username, password);
		fclose(fp);
	}
}

static void write_chap_secrets(Account *acct)
{
	FILE *fp;
	const char *username, *password;
  
	username = get_param_str(acct->section, "username", "");
	password = get_param_str(acct->section, "password", "");
	if(*username) {
		fp=fopen(CHAP_SECRETS_FILE, "w");

		fprintf(fp, "%s * \"%s\"\n", username, password);
		fprintf(fp, "* %s \"%s\"\n", username, password);
		fclose(fp);
	}
}
