#include <sys/time.h>
#include <picogui.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#include <fcntl.h>
#include <pguiwm.h>

#include "accounts.h"

static pghandle wLog, wStartStop;
static Account_list *al;
static int selectedAcct;
static char log_buf[1024];	

int list_widget(pghandle wInside);
int acct_widget(pghandle wInside);

int xtra_select(int n, fd_set *r, fd_set *w, fd_set *e, struct timeval *t);
void select_bh(int res, fd_set *r);

int setSelected (struct pgEvent *evt);
int pppStart (struct pgEvent *evt);
int appDone (struct pgEvent *evt);
int acctNew (struct pgEvent *evt);
int acctOpts (struct pgEvent *evt);
int acctDone (struct pgEvent *evt);
int irRestart (struct pgEvent *evt);

void crud_purge_log(char *buf, int stream);

static enum {LIST, ACCT, QUIT} display=LIST;

int main(int argc, char *argv[])
{
	pghandle wApp;
	pgInit(argc,argv);
	
	al = al_load();

	wApp = pgRegisterApp(PG_APP_NORMAL,"Network",0);
	wmRegisterApp(wApp, "Network", 0, NULL, NULL);
	
//	wNetwork=pgNewWidget(PG_WIDGET_BOX, 0, 0);

	pgCustomizeSelect(xtra_select, select_bh);

	do {
		pgEnterContext();
		if(display==LIST) list_widget(wApp);
		else acct_widget(wApp);
		pgEventLoop();
		pgLeaveContext();
	}while(display!=QUIT);

	wmUnregisterApp();

	return 0;
}

int list_widget(pghandle wInside)
{
	pghandle lastItem=0;
	int i;
	static int *pload = NULL;
	
	free(pload);
	pload=malloc(al->n_accts * sizeof(int));

	for(i=0;i<al->n_accts;i++) {
		lastItem = pgNewWidget (PG_WIDGET_LISTITEM,
					i ? PGDEFAULT : PG_DERIVE_INSIDE,
					i ? lastItem : wInside);
		if(acct_isconnected(al, i))
		        pgReplaceTextFmt(0, "%s <<connected>>",
					 acct_name(al, i));
		else
			pgSetWidget(0,
				    PG_WP_TEXT, pgNewString(acct_name(al, i)),
				    0);
		pload[i] = i;
		pgBind (0, PG_WE_ACTIVATE, &setSelected, pload+i);
		acct_sethandle(al, i, lastItem);
	}

	pgNewWidget(PG_WIDGET_TOOLBAR, i ? PGDEFAULT : PG_DERIVE_INSIDE,
		    i ? PGDEFAULT : wInside);
	pgNewWidget(PG_WIDGET_BUTTON, PG_DERIVE_INSIDE, 0);
	pgSetWidget(0,
		    PG_WP_TEXT, pgNewString("Done"),
		    0);
	pgBind(0, PG_WE_ACTIVATE, &appDone, 0);
	pgNewWidget(PG_WIDGET_BUTTON, 0, 0);
	pgSetWidget(0,
		    PG_WP_TEXT, pgNewString("New"),
		    0);
	pgBind(0, PG_WE_ACTIVATE, &acctNew, 0);
	return 0;
}

int acct_widget(pghandle wInside)
{
	pgNewWidget(PG_WIDGET_LABEL, PG_DERIVE_INSIDE, wInside);
	pgSetWidget(0,
		    PG_WP_TEXT, pgNewString(acct_name(al, selectedAcct)),
		    0);
	
	wLog = pgNewWidget(PG_WIDGET_LABEL, 0, 0);
	pgSetWidget(0, PG_WP_ALIGN, PG_A_LEFT, 0);
	pgReplaceText(wLog, log_buf);

	pgNewWidget(PG_WIDGET_TOOLBAR, 0, 0);
	pgSetWidget(0, PG_WP_SIDE, PG_S_BOTTOM, 0);

	wStartStop = pgNewWidget(PG_WIDGET_BUTTON, PG_DERIVE_INSIDE, 0);
	pgSetWidget(0,
		    PG_WP_TEXT, pgNewString(acct_isconnected(al, selectedAcct)
					    ? "Stop" : "Start"),
		    0);
	pgBind(0, PG_WE_ACTIVATE, &pppStart, 0);

	pgNewWidget(PG_WIDGET_BUTTON, 0, 0);
	pgSetWidget(0, PG_WP_TEXT, pgNewString("Opts"), 0);
	pgBind(0, PG_WE_ACTIVATE, &acctOpts, 0);

	pgNewWidget(PG_WIDGET_BUTTON, 0, 0);
	pgSetWidget(0, PG_WP_TEXT, pgNewString("Restart IR"), 0);
	pgBind(0, PG_WE_ACTIVATE, &irRestart, 0);

	pgNewWidget(PG_WIDGET_BUTTON, 0, 0);
	pgSetWidget(0, PG_WP_TEXT, pgNewString("Done"), 0);
	pgBind(0, PG_WE_ACTIVATE, &acctDone, 0);

	return 0;
}


int setSelected (struct pgEvent *evt)
{
	selectedAcct = *(int *)(evt->extra);

	display = ACCT;
	pgExitEventLoop();
	return 0;
}

int appDone (struct pgEvent *evt)
{
	display = QUIT;
	pgExitEventLoop();
	return 0;
}

int acctNew (struct pgEvent *evt)
{
	acct_edit(al, -1);
	display = LIST;
	pgExitEventLoop();
	return 0;
}

int acctOpts (struct pgEvent *evt)
{
	switch(pgMenuFromString("Edit|Delete")) {
	case 1:
		acct_edit(al, selectedAcct);
		display = ACCT;
		pgExitEventLoop();
		break;
	case 2:
		if(pgMessageDialogFmt("Delete Account",
				      PG_MSGBTN_YES | PG_MSGBTN_NO,
				      "Really delete account %s?",
				      acct_name(al, selectedAcct)) ==
		   PG_MSGBTN_YES) {
			al_delete_acct(al, selectedAcct);
			display = LIST;
			pgExitEventLoop();
			break;
		}
	}
	return 0;
}

int acctDone (struct pgEvent *evt)
{
	display = LIST;
	pgExitEventLoop();
	return 0;
}

int pppStart (struct pgEvent *evt)
{
	char buf[256];
	if(acct_isconnected(al, selectedAcct)) {
		sprintf(buf, "Stopping %s... ", acct_name(al, selectedAcct));
		write_log(buf);
		if(acct_stop(al, selectedAcct)==-1)
			write_log("failed\n");
		else
			write_log("done\n");
		pgReplaceText(wStartStop, "Start");
	} else {
		sprintf(buf, "Starting %s... ", acct_name(al, selectedAcct));
		write_log(buf);
		if(acct_start(al, selectedAcct)==-1)
			write_log("failed\n");
		else
			write_log("done\n");
		pgReplaceText(wStartStop, "Stop");
	}

	return 0;
}

int irRestart (struct pgEvent *evt)
{
	write_log("Stopping irattach... ");
	system("su -c \"/etc/rc.d/init.d/irattach stop\"");
	write_log("done\n");
	sleep(1);
	write_log("Restarting irattach... ");
	system("su -c \"/etc/rc.d/init.d/irattach start\"");
	write_log("done\n");
	return 0;
}

#define MAXLINES 20
void write_log(char *s)
{
	int i, n1=0, n2=0;
	char *p;

	for(p=log_buf;*p;p++) if(*p=='\n') n1++;
	for(p=s;*p;p++) if(*p=='\n') n2++;
	i=n1+n2-MAXLINES;
	p=log_buf;
	if(i>0) {
		for(p=log_buf;*p && i;p++) if(*p=='\n') i--;
	}
	i=strlen(p)+strlen(s)-1023;
	if(i>0) p+=i;
	memmove(log_buf, p, strlen(p)+1);
	strcat(log_buf, s);

	if(display == ACCT) {
		pgReplaceText(wLog, log_buf);
		pgUpdate();
	}
}

int xtra_select(int n, fd_set *r, fd_set *w, fd_set *e, struct timeval *t)
{
	int i;

	for(i=0;i<al->n_accts;i++) {
		if(acct_isconnected(al, i)) {
			if(al->accts[i].pfd>0)
				FD_SET(al->accts[i].pfd, r);
			if(al->accts[i].pfderr>0)
				FD_SET(al->accts[i].pfderr, r);
			if(al->accts[i].pfd >= n)
				n = al->accts[i].pfd + 1;
			if(al->accts[i].pfderr >= n)
				n = al->accts[i].pfderr + 1;
		}
	}
	return select(n, r, w, e, t);
}

void select_bh(int res, fd_set *r)
{
	int i, j;
	char buf[256];

	for(i=0;i<al->n_accts;i++) {
		if(al->accts[i].pfd>0 &&
		   FD_ISSET(al->accts[i].pfd, r)) {
			j = read(al->accts[i].pfd, buf, 255);
			if(j<1) 
				al->accts[i].pfd=-1;
			else {
				buf[j]=0;
				crud_purge_log(buf, 0);
			}
		}
		if(al->accts[i].pfderr>0 &&
		   FD_ISSET(al->accts[i].pfderr, r)) {
			j = read(al->accts[i].pfderr, buf, 255);
			if(j<1)
				al->accts[i].pfd=-1;
			else {
				buf[j]=0;
				crud_purge_log(buf, 1);
			}
		}
	}
}

void crud_purge_log(char *buf, int stream)
{
	static char line0[256]={0}, line1[256]={0}, *line;
	char *p;
	
	line=stream ? line1 : line0;

	for(;;) {
		int i=0;
		for(p=line + strlen(line);*buf;buf++){ 
			if(++i > 255) break;
			*p++=*buf;
			if(*buf=='\n') break;
		}
		*p=0;
		if(!*buf++) return; /*Incomplete line*/
		if(strncmp(line, "sent", 4) && strncmp(line, "rcvd", 4))
			write_log(line);
		*line=0;
	}
}
